import click
import xarray as xr
import multiprocessing as mp
from pathlib import Path
from functools import partial
from typing import Optional

from .shared import open_yaml_file, default_config_file
from .adas_interface.download_adas_datasets import download_species_data
from .read_rate_coeffs import read_rate_coeff

from .coronal_equilibrium import calculate_coronal_fractional_abundances
from .radiated_power import calculate_Lz
from .time_evolution import calculate_time_evolution
from .unit_handling import convert_units, ureg
from .mavrin_reference import compare_radas_to_mavrin


@click.command()
@click.option(
    "-d",
    "--directory",
    type=click.Path(),
    default=Path("./radas_dir").absolute(),
    help="Directory for storing work files and outputs. DEFAULT: ./radas_dir",
)
@click.option(
    "-c",
    "--config",
    type=click.Path(exists=True),
    default=None,
    help="Path to a yaml file for configuring radas. DEFAULT: RADAS_DIR/radas/config.yaml",
)
@click.option(
    "-s",
    "--species",
    default=("all",),
    multiple=True,
    help="Species to perform analysis for ('species_name'|'all'|'none'). DEFAULT: all",
)
@click.option(
    "-v", "--verbose", count=True, help="Write additional output to the command line."
)
@click.option(
    "--debug",
    is_flag=True,
    help="Flag to enable debug mode (disables multiprocessing).",
)
def run_radas_cli(
    directory: Path,
    config: Optional[str],
    species: list[str],
    verbose: int,
    debug: bool,
):
    """Runs the radas program.

    If config is given, it must point to a config.yaml file. Otherwise, a
    default config.yaml (stored in the module directory) is used.

    If species is given, it must be a valid species name (i.e. 'hydrogen').
    Otherwise, all valid species in the config.yaml file are evaluated.
    """
    kwargs = dict(
        directory=directory,
        config=config,
        species=species,
        verbose=verbose,
        debug=debug,
    )
    try:
        from ipdb import launch_ipdb_on_exception

        with launch_ipdb_on_exception():
            run_radas(**kwargs)
    except ModuleNotFoundError:
        run_radas(**kwargs)


def run_radas(
    directory: Path,
    config: Optional[str],
    species: list[str],
    verbose: int,
    debug: bool,
):

    radas_dir = Path(directory)
    if verbose:
        print(f"Running radas in {radas_dir.absolute()}")
    data_file_dir = radas_dir / "data_files"
    output_dir = radas_dir / "output"

    for path in [radas_dir, data_file_dir, output_dir]:
        path.mkdir(exist_ok=True)

    if species == ("none",):
        if verbose:
            print("Skipping computation.")
    else:
        config_file = default_config_file if config is None else Path(config).absolute()
        if verbose:
            print(f"Opening config file at {config_file}")
        configuration = open_yaml_file(config_file)

        if verbose:
            print(f"Downloading data from OpenADAS to {data_file_dir.absolute()}")
        for species_name, species_config in configuration["species"].items():
            if "data_files" in species_config and (
                (species_name in species) or (species == ("all",))
            ):
                download_species_data(
                    data_file_dir,
                    species_name,
                    species_config,
                    configuration["data_file_config"],
                    verbose=verbose,
                )

        if verbose:
            print("Reading rate coefficients")
        datasets = dict()
        for species_name, species_config in configuration["species"].items():
            if "data_files" in species_config and (
                (species_name in species) or (species == ("all",))
            ):
                datasets[species_name] = read_rate_coeff(
                    data_file_dir, species_name, configuration
                )

        output_dir.mkdir(exist_ok=True, parents=True)
        if not debug:
            with mp.Pool() as pool:
                if species != ("all",):
                    datasets = {
                        species_name: datasets[species_name] for species_name in species
                    }

                pool.map(
                    partial(
                        run_radas_computation, output_dir=output_dir, verbose=verbose
                    ),
                    [(ds) for ds in datasets.values()],
                )
        else:
            for ds in datasets.values():
                run_radas_computation(ds, output_dir=output_dir, verbose=verbose)

    if verbose:
        print(f"Generating plots and saving output to {output_dir}")
    compare_radas_to_mavrin(output_dir)

    if verbose:
        print("Done")


def run_radas_computation(dataset: xr.Dataset, output_dir: Path, verbose: int):
    """Calculate several dependent quantities based on the atomic rates, and store
    the result as a NetCDF file.
    """
    if verbose:
        print(f"Running computation for {dataset.species_name}")

    dataset["coronal_charge_state_fraction"] = calculate_coronal_fractional_abundances(
        dataset
    )
    dataset["coronal_mean_charge_state"] = (
        dataset.coronal_charge_state_fraction * dataset.dim_charge_state
    ).sum(dim="dim_charge_state")
    dataset["coronal_Lz"] = calculate_Lz(dataset, dataset.coronal_charge_state_fraction)
    dataset["residence_time"] = convert_units(
        dataset.ne_tau / dataset.electron_density, ureg.s
    )
    dataset["charge_state_evolution"] = calculate_time_evolution(dataset)
    dataset["equilibrium_charge_state_fraction"] = dataset.charge_state_evolution.isel(
        dim_time=-1
    )
    dataset["equilibrium_mean_charge_state"] = (
        dataset.equilibrium_charge_state_fraction * dataset.dim_charge_state
    ).sum(dim="dim_charge_state")
    dataset["equilibrium_Lz"] = calculate_Lz(
        dataset, dataset.equilibrium_charge_state_fraction
    )

    output_dir.mkdir(exist_ok=True)
    dataset.pint.dequantify().to_netcdf(output_dir / f"{dataset.species_name}.nc")


@click.command()
@click.option(
    "-o",
    "--output",
    type=click.Path(),
    default=Path("./config.yaml").absolute(),
    help="Output path for a config file. DEFAULT: ./config.yaml",
)
def write_config_template(output: Path):
    output = Path(output).absolute()
    print(f"Copying {default_config_file} to {output}")
    output.write_text(default_config_file.read_text())
    print("Done")



from .compare_to_mavrin import (
    compare_radas_to_mavrin,
    compare_radas_to_mavrin_per_species,
    compute_Mavrin_polynomial_fit,
)
from .read_mavrin_data import read_mavrin_data

__all__ = [
    "compare_radas_to_mavrin",
    "compare_radas_to_mavrin_per_species",
    "compute_Mavrin_polynomial_fit",
    "read_mavrin_data",
]



def determine_reader_class_and_config(data_file_config, dataset_type):
    """Examines the data_file_config to determine which reader class to use to read a specific dataset_type."""
    for reader_key, reader_config in data_file_config.items():
        for dataset_key, dataset_config in reader_config.items():
            if dataset_key == dataset_type:
                return reader_key, dataset_config

    raise NotImplementedError(f"Cannot identify reader for {dataset_type}.")



import numpy as np
import xarray as xr
from scipy.integrate import solve_ivp
from .unit_handling import ureg, magnitude_in_units


def calculate_time_evolution(dataset: xr.Dataset) -> xr.Dataset:
    """Evolve the system over time, and record the impurity charge-state fractions as a function of time.

    The equations are stiff, so we need to use "BDF", "Radau" or "LSODA" as the solver method. Radau was
    found to give a good balance of accuracy and speed.
    """
    evaluation_times = np.logspace(
        np.log10(magnitude_in_units(dataset.evolution_start, ureg.s)).item(),
        np.log10(magnitude_in_units(dataset.evolution_stop, ureg.s)).item(),
    )

    def _time_evolve(
        effective_ionisation,
        effective_recombination,
        electron_density,
        ne_tau,
    ):
        charge_state_fraction = np.zeros_like(effective_ionisation)
        charge_state_fraction[0] = 1.0

        result = solve_ivp(
            calculate_derivative,
            y0=charge_state_fraction,
            t_span=[evaluation_times[0], evaluation_times[-1]],
            t_eval=evaluation_times,
            args=(
                effective_ionisation,
                effective_recombination,
                electron_density,
                ne_tau,
            ),
            method="Radau",
            rtol=1e-3,
            atol=1e-12,
        )

        return result.y

    charge_state_fraction = xr.apply_ufunc(
        _time_evolve,
        magnitude_in_units(dataset.effective_ionisation, ureg.m**3 / ureg.s),
        magnitude_in_units(
            dataset.effective_recombination.roll(dim_charge_state=-1),
            ureg.m**3 / ureg.s,
        ),
        magnitude_in_units(dataset.electron_density, ureg.m**-3),
        magnitude_in_units(dataset.ne_tau, ureg.m**-3 * ureg.s),
        vectorize=True,
        input_core_dims=[("dim_charge_state",), ("dim_charge_state",), (), ()],
        output_core_dims=[("dim_charge_state", "dim_time")],
    ).assign_coords(dim_time=evaluation_times)

    return charge_state_fraction


def shift(arr, num, fill_value=0.0):
    result = np.empty_like(arr)
    if num > 0:
        result[:num] = fill_value
        result[num:] = arr[:-num]
    elif num < 0:
        result[num:] = fill_value
        result[:num] = arr[-num:]
    else:
        result[:] = arr
    return result


def calculate_derivative(
    _,
    charge_state_fraction: np.ndarray,
    effective_ionisation,
    effective_recombination,
    electron_density,
    ne_tau=np.inf,
):
    """Calculate the the partial derivative of the impurity density w.r.t. time.

    A compensated sum is used to reduce the risk of truncation. If you think this is affecting performance, you can replace
    the last lines with
    dydt = np.sum([-ionisation_to_above, ionisation_from_below, recombination_from_above, -recombination_to_below], axis=0)

    The default option is for a non-refuelled impurity. If you set a ne_tau, it is assumed that the ground state is
    constantly refuelled at a rate of 1 / ne_tau and that the excited states
    are lost at a rate proportional to their concentration.
    """
    ionisation_to_above = effective_ionisation * charge_state_fraction
    ionisation_from_below = shift(effective_ionisation * charge_state_fraction, +1)

    recombination_from_above = effective_recombination * shift(
        charge_state_fraction, -1
    )
    recombination_to_below = shift(effective_recombination, +1) * charge_state_fraction

    change_in_charge_state_fraction = np.zeros_like(charge_state_fraction)
    for i in range(len(change_in_charge_state_fraction)):
        change_in_charge_state_fraction[i] = kahan_babushka_neumaier_sum(
            [
                -ionisation_to_above[i],
                ionisation_from_below[i],
                recombination_from_above[i],
                -recombination_to_below[i],
            ]
        )

    change_in_charge_state_fraction -= charge_state_fraction / ne_tau
    change_in_charge_state_fraction[0] += 1.0 / ne_tau

    return change_in_charge_state_fraction * electron_density


def kahan_babushka_neumaier_sum(values_to_sum):
    """Improved Kahan compensated summation algorithm."""
    running_sum = 0.0
    compensation = 0.0

    for value in values_to_sum:
        temporary_sum = running_sum + value

        if running_sum >= value:
            compensation += running_sum - temporary_sum + value
        else:
            compensation += value - temporary_sum + running_sum

        running_sum = temporary_sum

    return running_sum + compensation



"""Bits and pieces shared by many other parts of radas."""

from pathlib import Path
from importlib.resources import files
import subprocess
import yaml

default_config_file = files("radas").joinpath("config.yaml")
mavrin_data_file = files("radas.mavrin_reference").joinpath("mavrin_data.yaml")

library_extensions = [".a", ".so"]


def get_git_revision_short_hash() -> str:
    try:
        return (
            subprocess.check_output(["git", "rev-parse", "--short", "HEAD"])
            .decode("ascii")
            .strip()
        )
    except: # noqa:E722
        # If git isn't available (sometimes the case in tests), return a blank
        return ""


def open_yaml_file(yaml_file: Path) -> dict:
    with open(yaml_file, "r") as file:
        return yaml.load(file, Loader=yaml.FullLoader)



import xarray as xr
from radas.unit_handling import convert_units, ureg


def calculate_Lz(
    dataset: xr.Dataset, charge_state_fraction: xr.DataArray
) -> xr.DataArray:
    """Calculate the radiated power prefactor due to electron-impurity interactions.

    You can calculate P_rad_electron = n_e * n_z * electron_emission
    where n_z = impurity density summed over charge states
    """

    electron_emission_per_charge_state = (
        dataset.line_emission_from_excitation + dataset.recombination_and_bremsstrahlung
    ) * charge_state_fraction
    electron_emission = electron_emission_per_charge_state.sum(dim="dim_charge_state")

    return convert_units(electron_emission, ureg.W * ureg.m**3)



"""Set up the pint library for unit handling."""

from typing import Any, Union

import numpy as np
import numpy.typing as npt
import pint
import pint_xarray
import xarray as xr
from pint import UnitStrippedWarning, DimensionalityError

ureg = pint_xarray.setup_registry(
    pint.UnitRegistry(
        force_ndarray_like=True,
    )
)
pint.set_application_registry(ureg)

Quantity = ureg.Quantity

def convert_units(
    array: Union[xr.DataArray, pint.Quantity], units: Any
) -> Union[xr.DataArray, pint.Quantity]:
    """Convert an array to specified units, handling both Quantities and xr.DataArrays."""
    if isinstance(array, xr.DataArray):
        if not hasattr(array.pint, "units") or array.pint.units is None:
            array = array.pint.quantify(ureg.dimensionless)

        return array.pint.to(units)  # type: ignore[no-any-return]
    elif isinstance(array, Quantity):
        return array.to(units)  # type:ignore[no-any-return]
    else:
        raise NotImplementedError(
            f"No implementation for 'convert_units' with an array of type {type(array)} ({array})"
        )


def magnitude(
    array: Union[xr.DataArray, pint.Quantity]
) -> Union[npt.NDArray[np.float32], float]:
    """Return the magnitude of an array, handling both Quantities and xr.DataArrays."""
    if isinstance(array, xr.DataArray):
        return array.pint.dequantify()
    elif isinstance(array, Quantity):
        return array.magnitude  # type: ignore[no-any-return]
    else:
        raise NotImplementedError(
            f"No implementation for 'magnitude' with an array of type {type(array)} ({array})"
        )


def dimensionless_magnitude(
    array: Union[xr.DataArray, pint.Quantity]
) -> Union[npt.NDArray[np.float32], float]:
    """Converts the array to dimensionless and returns the magnitude."""
    return magnitude(convert_units(array, ureg.dimensionless))


def magnitude_in_units(
    array: Union[xr.DataArray, pint.Quantity], units: Any
) -> Union[npt.NDArray[np.float32], float]:
    """Convert the array to the specified units and then return the magnitude."""
    return magnitude(convert_units(array, units))


__all__ = [
    "DimensionalityError",
    "UnitStrippedWarning",
    "ureg",
    "Quantity",
    "convert_units",
    "magnitude",
    "dimensionless_magnitude",
]



from .unit_handling import (
    DimensionalityError,
    UnitStrippedWarning,
    ureg,
    Quantity,
    convert_units,
    magnitude,
    dimensionless_magnitude,
)

from .cli import (
    run_radas_cli,
    run_radas,
    run_radas_computation,
    write_config_template,
)

from .read_rate_coeffs import read_rate_coeff
from .coronal_equilibrium import calculate_coronal_fractional_abundances
from .radiated_power import calculate_Lz
from .time_evolution import calculate_time_evolution

__all__ = [
    "DimensionalityError",
    "UnitStrippedWarning",
    "ureg",
    "Quantity",
    "convert_units",
    "magnitude",
    "dimensionless_magnitude",
    "run_radas_cli",
    "run_radas",
    "run_radas_computation",
    "read_rate_coeff",
    "calculate_coronal_fractional_abundances",
    "calculate_Lz",
    "calculate_time_evolution",
    "write_config_template",
]



from .unit_handling import Quantity, ureg, convert_units, dimensionless_magnitude
from .adas_interface.determine_adas_dataset_type import (
    determine_reader_class_and_config,
)
from .shared import get_git_revision_short_hash
import xarray as xr
import numpy as np


def read_rate_coeff(data_file_dir, species_name, config):
    """Builds a rate_dataset combining all of the raw data available for a given species."""
    config_for_species = config["species"][species_name]

    dataset = xr.Dataset().assign_attrs(
        atomic_number=config_for_species["atomic_number"],
        species_name=species_name,
        git_hash=get_git_revision_short_hash(),
    )

    dataset = write_global_attributes(dataset, config["globals"])

    for dataset_type in config_for_species["data_files"].keys():
        reader_key, dataset_config = determine_reader_class_and_config(
            config["data_file_config"], dataset_type
        )

        if reader_key == "adf11":
            rate_dataset = build_adf11_rate_dataset(
                data_file_dir,
                species_name,
                dataset_type,
                dataset_config,
            )
        else:
            raise NotImplementedError(
                f"No implementation for reading {reader_key} files."
            )

        determine_coordinates(dataset, rate_dataset)
        dataset[dataset_type] = rate_dataset.rate_coefficient

    dataset = align_rates_on_charge_states(dataset)

    return dataset


def write_global_attributes(dataset: xr.Dataset, globals: dict) -> xr.Dataset:
    for attribute, value in globals.items():
        if isinstance(value, dict):
            if np.size(value["value"]) > 1:
                dataset[attribute] = xr.DataArray(
                    Quantity(value["value"], value["units"]),
                    coords={f"dim_{attribute}": value["value"]},
                )
            else:
                dataset[attribute] = Quantity(value["value"], value["units"])
        else:
            dataset[attribute] = value

    return dataset


def determine_coordinates(dataset: xr.Dataset, rate_dataset: xr.Dataset):

    for key in [
        "electron_density",
        "electron_temp",
        "reference_electron_density",
        "reference_electron_temp",
    ]:
        if key not in dataset:
            dataset[key] = rate_dataset[key]

    for key in ["electron_density", "electron_temp"]:
        np.testing.assert_allclose(
            dimensionless_magnitude(
                (dataset[key] - rate_dataset[key]) / dataset[f"reference_{key}"]
            ),
            0.0,
        )


def build_adf11_rate_dataset(
    data_file_dir, species_name, dataset_type, dataset_config
):
    from .adas_interface.read_adf11_file import read_adf11_file

    data = read_adf11_file(data_file_dir, species_name, dataset_type)

    ds = xr.Dataset()

    ds["species"] = species_name
    ds["dataset"] = dataset_type
    ds["charge"] = data["IZMAX"]

    electron_density = convert_units(
        Quantity(10 ** data["DDENSD"][: data["IDMAXD"]], ureg.cm**-3), ureg.m**-3
    )
    electron_temp = Quantity(10 ** data["DTEVD"][: data["ITMAXD"]], ureg.eV)

    # Use logarithmic quantities to define the coordinates, so that we can interpolate over logarithmic quantities.
    ds["electron_density"] = xr.DataArray(
        electron_density, coords=dict(dim_electron_density=electron_density.magnitude)
    )
    ds["electron_temp"] = xr.DataArray(
        electron_temp, coords=dict(dim_electron_temp=electron_temp.magnitude)
    )

    ds["reference_electron_density"] = Quantity(1.0, ureg.m**-3)
    ds["reference_electron_temp"] = Quantity(1.0, ureg.eV)

    ds["number_of_charge_states"] = data["IZMAX"]
    charge_state = np.arange(data["IZMAX"])
    ds["charge_state"] = xr.DataArray(
        charge_state, coords=dict(dim_charge_state=charge_state)
    )

    coefficient = data["DRCOFD"][: data["IZMAX"], : data["ITMAXD"], : data["IDMAXD"]]
    if dataset_config["code"] <= 9:
        coefficient = 10**coefficient

    input_units = dataset_config["stored_units"]
    output_units = dataset_config["desired_units"]
    ds["rate_coefficient"] = convert_units(
        xr.DataArray(
            coefficient,
            dims=("dim_charge_state", "dim_electron_temp", "dim_electron_density"),
        ).pint.quantify(input_units),
        output_units,
    )

    return ds


def align_rates_on_charge_states(dataset: xr.Dataset) -> xr.Dataset:
    """For rates which are for k+1->k reactions, we shift these by one position
    in the dim_charge_state dimension so that the kth position of a rate always
    refers to the reactant species."""

    dataset = dataset.pad(
        pad_width=dict(dim_charge_state=(0, 1)), mode="constant", constant_values=0.0
    )
    dataset = dataset.assign_coords(
        dim_charge_state=np.arange(dataset.sizes["dim_charge_state"])
    )

    for key in [
        "effective_recombination",
        "charge_exchange_cross_coupling",
        "recombination_and_bremsstrahlung",
        "charge_exchange_emission",
    ]:
        dataset[key] = dataset[key].roll(dim_charge_state=+1)

    return dataset



import numpy as np
from fortranformat import FortranRecordReader as FReader

def read_single_values(list_of_lines, format_spec, start_at_line):
    """
    Makes a 'generator', which is a special Python object that returns a value each time that it is called.

    This generator reads each line in the list of lines, and tries to format it according to format_spec
    For instance, the default format '5e16.9' will try to read 5 fortran-formatted floats in exponential format
    with a width of 16 characters and 9 characters after the decimal point
    Where there are less than 5 numbers on a line, the fortran reader will return 'None'. These are skipped (not
    returned by the generator)

    The generator returns the next value according to the specified format, and also the line number and position being read
    """
    for line_number, line in enumerate(list_of_lines):
        if line_number < start_at_line:
            continue
        
        try:
            values = FReader(f"({format_spec})").read(line)
        except ValueError as e:
            message = (
                f"fortranformat could not read line {line_number}: {line} with format ({format_spec})\n"
                "Make sure that your format_spec has the correct number of elements, the correct spacing and the correct format for the line.\n"
                "See https://docs.oracle.com/cd/E19957-01/805-4939/6j4m0vn9f/index.html for reference.\n"
                f"fortranformat error was: {e}"
            )
            raise ValueError(message) from e
            

        for value_position, value in enumerate(values):
            if value is not None:
                yield value, line_number, value_position

def read_1d_array(generator, number_of_values):
    """
    Take number_of_values from the generator and write it into a 1D array
    """
    array = np.zeros(number_of_values)
    line_number, value_position = 0, 0

    for i in range(number_of_values):
        value, line_number, value_position = next(generator)

        if i == 0:
            assert value_position == 0, \
                "Arrays are always defined as starting on a new line, but the first element in this array is not at position 0"

        array[i] = value

    return array, line_number, value_position

def read_2d_array(generator, rows, columns, fortran_order_arrays=False):
    """
    Take rows*columns from the generator and write it into a 2D array of
    shape (rows, columns), or (columns, rows) if transpose = True
    """
    number_of_values = rows * columns
    array, line_number, value_position = read_1d_array(generator, number_of_values)

    if fortran_order_arrays:
        array = array.reshape((rows, columns))
    else:
        array = array.reshape((columns, rows))

    return array, line_number, value_position



import xarray as xr
import numpy as np
import warnings
from ..unit_handling import ureg, convert_units, magnitude


def read_mavrin_data():
    from ..shared import open_yaml_file, mavrin_data_file

    "Opens the Mavrin dataset from a yaml file."
    return open_yaml_file(mavrin_data_file)


def compute_Mavrin_polynomial_fit(Te_eV, ne_tau_s_per_m3, coeff):
    """Compute Lz or mean_charge curves from Mavrin, J. Fus. Eng., 2017."""
    return xr.apply_ufunc(
        compute_Mavrin_polynomial_fit_single,
        magnitude(convert_units(Te_eV, ureg.eV)),
        magnitude(convert_units(ne_tau_s_per_m3, ureg.s / ureg.m**3)),
        kwargs=dict(coeff=coeff),
        vectorize=True,
    )


def compute_Mavrin_polynomial_fit_single(
    Te_eV, ne_tau_s_per_m3, coeff, warn: bool = False
):
    """Inner loop for computing the Lz or mean_charge polynomial fit from Mavrin, J. Fus. Eng., 2017."""
    Tmin_eV = coeff["Tmin_eV"]
    Tmax_eV = coeff["Tmax_eV"]

    if not Tmin_eV[0] <= Te_eV <= Tmax_eV[-1]:
        if warn:
            warnings.warn(
                f"{Te_eV}eV outside fitted range {Tmin_eV[0]}eV to {Tmax_eV[-1]}eV"
            )
        return np.nan
    if ne_tau_s_per_m3 < 1e15:
        if warn:
            warnings.warn(f"{ne_tau_s_per_m3} outside fitted range above 1e16 m^-3 s")
        return np.nan

    X = np.log10(Te_eV)
    Y = np.log10(ne_tau_s_per_m3 / 1e19)
    if warn and (Y > 0.0):
        warnings.warn(
            "Warning: treating points with ne_tau_s_per_m3 > 1e19 m^-3 s as coronal."
        )
    Y = np.minimum(Y, 0.0)

    N_bins = len(Tmin_eV)
    assert len(Tmax_eV) == N_bins

    for i in range(N_bins):
        if Tmin_eV[i] <= Te_eV <= Tmax_eV[i]:
            T_bin = i

    A = np.zeros(10)
    for i in range(10):
        A[i] = coeff[f"A{i}"][T_bin]

    F = (
        A[0]
        + A[1] * X
        + A[2] * Y
        + A[3] * X**2
        + A[4] * X * Y
        + A[5] * Y**2
        + A[6] * X**3
        + A[7] * X**2 * Y
        + A[8] * X * Y**2
        + A[9] * Y**3
    )

    return np.power(10, F)



import xarray as xr
from pathlib import Path
import matplotlib.pyplot as plt
from .read_mavrin_data import (
    read_mavrin_data,
    compute_Mavrin_polynomial_fit,
)
from ..unit_handling import ureg


def compare_radas_to_mavrin(output_dir: Path):

    for output_file in output_dir.iterdir():
        if output_file.suffix == ".nc":
            species = output_file.stem
            compare_radas_to_mavrin_per_species(output_dir, species)


def compare_radas_to_mavrin_per_species(output_dir: Path, species: str):
    mavrin_data = read_mavrin_data()

    ds = xr.open_dataset(output_dir / f"{species}.nc").pint.quantify()
    ds = ds.sel(dim_electron_density=1e20, method="nearest")

    Te = ds["electron_temp"]
    ne_tau = ds["ne_tau"]

    if f"{species}_Lz" in mavrin_data:
        Lz_coeffs = mavrin_data.get(f"{species}_Lz")
        Lz_mavrin = (
            compute_Mavrin_polynomial_fit(Te, ne_tau, coeff=Lz_coeffs)
            .squeeze()
            .pint.quantify(ureg.W * ureg.m**3)
        )
    else:
        Lz_mavrin = None

    if f"{species}_mean_charge" in mavrin_data:
        mean_charge_coeffs = mavrin_data.get(f"{species}_mean_charge")
        mean_charge_mavrin = compute_Mavrin_polynomial_fit(
            Te, ne_tau, coeff=mean_charge_coeffs
        ).squeeze()
    else:
        mean_charge_mavrin = None

    Lz_radas = ds["equilibrium_Lz"].pint.to(ureg.W * ureg.m**3)
    mean_charge_radas = ds["equilibrium_mean_charge_state"].pint.to(ureg.dimensionless)

    fig, axs = plt.subplots(ncols=2)

    for i in range(ds.sizes["dim_ne_tau"]):
        ne_tau = ds.ne_tau.isel(dim_ne_tau=i).item()

        Lz_radas.isel(dim_ne_tau=i).plot(ax=axs[0], label=f"{ne_tau:~P}", color=f"C{i}")
        if Lz_mavrin is not None:
            Lz_mavrin.isel(dim_ne_tau=i).plot(ax=axs[0], color=f"C{i}", linestyle="--")

        mean_charge_radas.isel(dim_ne_tau=i).plot(ax=axs[1], color=f"C{i}")
        if mean_charge_mavrin is not None:
            mean_charge_mavrin.isel(dim_ne_tau=i).plot(
                ax=axs[1], color=f"C{i}", linestyle="--"
            )

    axs[0].legend()
    axs[0].set_yscale("log")
    if Lz_mavrin is not None:
        axs[0].set_ylim(*Lz_coeffs["ylims"])
    if mean_charge_mavrin is not None:
        axs[1].set_ylim(*mean_charge_coeffs["ylims"])

    axs[0].set_title("$L_z$ $[W m^3]$")
    axs[1].set_title("$<Z>$")

    for ax in axs.flatten():
        ax.set_xscale("log")
        ax.set_xlabel("$T_e$ [$eV$]")
        ax.set_ylabel("")

    plt.suptitle(species)

    plt.savefig(output_dir / f"{species}.png")



from .determine_adas_dataset_type import determine_reader_class_and_config
from .download_adas_datasets import download_species_data
from .read_adf11_file import read_adf11_file

__all__ = [
    "determine_reader_class_and_config",
    "download_species_data",
    "read_adf11_file",
]



import xarray as xr
from .unit_handling import dimensionless_magnitude


def calculate_coronal_fractional_abundances(dataset: xr.Dataset) -> xr.DataArray:
    """Calculate the fractional abundances of different charge states, assuming coronal equilibrium."""

    ratio_of_ionisation_to_recombination = (
        dataset.effective_ionisation
        / dataset.effective_recombination.roll(dim_charge_state=-1)
    )
    charge_state_fraction = xr.zeros_like(ratio_of_ionisation_to_recombination)
    ratio_of_ionisation_to_recombination = dimensionless_magnitude(
        ratio_of_ionisation_to_recombination
    )

    charge_state_fraction[0] = 1.0

    for charge_state in range(dataset.atomic_number):
        charge_state_fraction[charge_state + 1] = (
            charge_state_fraction[charge_state]
            * ratio_of_ionisation_to_recombination[charge_state]
        )

    charge_state_fraction /= charge_state_fraction.sum(dim="dim_charge_state")

    return charge_state_fraction



from pathlib import Path
import importlib.util
from .fortran_reader import FReader, read_single_values, read_1d_array, read_2d_array
import numpy as np


def load_library(library_name: str, filepath: Path):
    spec = importlib.util.spec_from_file_location(library_name, filepath)

    library = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(library)

    return library


def read_adf11_file(
    data_file_dir, species_name, dataset_type
) -> dict:
    """Open and read an ADF11 OpenADAS file.
    
    Uses the format specification from https://www.adas.ac.uk/man/appxa-11.pdf
    """


    filename = data_file_dir / f"{species_name}_{dataset_type}.dat"
    if not filename.exists():
        raise FileNotFoundError(f"{filename} does not exist.")

    f = filename.read_text().split("\n")

    IZMAX, IDMAXD, ITMAXD, IZ1MIN, IZ1MAX = FReader("5i5").read(f[0])

    # Initialise the value reader
    
    # IMPORTANT: do not reinitialise the reader, since otherwise we'll loose the position of the reader
    # Note that line_numbers will be reported from the read-offset
    value_reader = read_single_values(f, start_at_line=2, format_spec="8f10.5")

    DDENSD, line_number, _ = read_1d_array(value_reader, number_of_values=IDMAXD)
    DTEVD, line_number, _ = read_1d_array(value_reader, number_of_values=ITMAXD)

    DRCOFD = np.zeros((IZMAX, ITMAXD, IDMAXD))

    for IZ1 in range(IZMAX):
        value_reader = read_single_values(f, start_at_line=line_number + 2, format_spec="8f10.5")
        DRCOFD[IZ1, :, :], line_number, _ = read_2d_array(value_reader, rows=IDMAXD, columns=ITMAXD, fortran_order_arrays=False)

    data = dict(
        IZMAX = IZMAX,
        IDMAXD = IDMAXD,
        ITMAXD = ITMAXD,
        IZ1MIN = IZ1MIN,
        IZ1MAX = IZ1MAX,
        DDENSD = DDENSD,
        DTEVD = DTEVD,
        DRCOFD = DRCOFD,
    )

    return data



from pathlib import Path
import urllib.request
from .determine_adas_dataset_type import determine_reader_class_and_config


def download_species_data(
    data_file_dir: Path,
    species_name: str,
    species_config: dict,
    data_file_config: dict,
    verbose: int,
    url_base: str = "https://open.adas.ac.uk",
):
    """Downloads all of the data files for a specific species."""
    data_file_dir.mkdir(exist_ok=True, parents=True)

    for dataset_type, file_to_download in species_config["data_files"].items():

        reader_class, dataset_config = determine_reader_class_and_config(
            data_file_config, dataset_type
        )

        if isinstance(file_to_download, int):
            species_key = species_config["atomic_symbol"].lower()
            year = file_to_download
        elif isinstance(file_to_download, list) and len(file_to_download) == 2:
            species_key = file_to_download[0].lower()
            year = file_to_download[1]
        else:
            raise NotImplementedError(f"Could not process entry: {file_to_download} for {species_name} {dataset_type}")

        year_key = f"{year}"[-2:]
        dataset_prefix = dataset_config["prefix"].lower()

        output_filename = data_file_dir / f"{species_name}_{dataset_type}.dat"
        query_path = f"{url_base}/download/{reader_class}/{dataset_prefix}{year_key}/{dataset_prefix}{year_key}_{species_key}.dat"

        if not output_filename.exists():
            if verbose >= 2:
                print(f"Downloading {query_path} to {output_filename}")
            urllib.request.urlretrieve(query_path, output_filename)
        else:
            if verbose >= 2:
                print(f"Reusing {query_path} ({output_filename} already exists)")

        if "OPEN-ADAS Error" in output_filename.read_text():
            output_filename.unlink()
            if verbose:
                print(
                    f"Failed to download the {year} {dataset_prefix} for {species_name}"
                )
